from .constants import Serializer
from .myjson import MyJson
from .myxml import MyXml


class Factory:

    @staticmethod
    def create_serializer(serializer: Serializer):
        if Serializer.Json:
            return MyJson()

        elif Serializer.Xml:
            return MyXml()



from .constants import CODE_PROPERTIES, UNIQUE_TYPES, TYPE, SOURCE, CODE,\
    GLOBALS, NAME, DEFAULTS, CLOSURE, BASES, DICT, CLASS, OBJECT
import inspect
from types import NoneType, ModuleType, CodeType, FunctionType, CellType


class ObjectConverter:
    @classmethod
    def get_dict(cls, obj, is_inner_func=False):
        if type(obj) in (int, float, bool, str, NoneType):
            return obj

        if type(obj) is list:
            return [cls.get_dict(i) for i in obj]

        result = {TYPE: type(obj).__name__}

        if type(obj) is dict:
            result[SOURCE] = [[cls.get_dict(key), cls.get_dict(value)] for key, value in obj.items()]
            return result

        if type(obj) in (set, frozenset, tuple, bytes, bytearray):
            result[SOURCE] = cls.get_dict([*obj])
            return result

        if type(obj) is complex:
            result[SOURCE] = {complex.real.__name__: obj.real,
                              complex.imag.__name__: obj.imag}
            return result

        if type(obj) is ModuleType:
            # result[SOURCE] = obj.__name__
            # return result
            return {TYPE: ModuleType.__name__,
                    SOURCE: obj.__name__}

        if type(obj) is CodeType:
            result[SOURCE] = cls.__get_code_dict(obj)
            return result

        if type(obj) is CellType:
            result[SOURCE] = cls.get_dict(obj.cell_contents)
            return result

        if type(obj) in (staticmethod, classmethod):
            result[SOURCE] = cls.get_dict(obj.__func__, is_inner_func)
            return result

        if inspect.isroutine(obj):
            result[SOURCE] = cls.__get_routine_dict(obj, is_inner_func)
            return result

        elif inspect.isclass(obj):
            result[SOURCE] = cls.__get_class_dict(obj)
            return result

        else:
            result[SOURCE] = cls.__get_object_dict(obj)
            return result

    @classmethod
    def get_object(cls, dictionary, is_dict=False):
        if is_dict:
            return {cls.get_object(item[0]): cls.get_object(item[1]) for item in dictionary}

        if type(dictionary) not in (dict, list):
            return dictionary

        elif type(dictionary) is list:
            return [cls.get_object(o) for o in dictionary]

        else:
            object_type = dictionary[TYPE]
            object_source = dictionary[SOURCE]

            if object_type == dict.__name__:
                return cls.get_object(object_source, is_dict=True)

            cols_dict = {t.__name__: t for t in [
                set, frozenset, tuple, bytes, bytearray]}

            if object_type in cols_dict:
                return cols_dict[object_type](cls.get_object(object_source))

            if object_type == complex.__name__:
                return object_source[complex.real.__name__] + object_source[complex.imag.__name__] * 1j

            if object_type == ModuleType.__name__:
                return __import__(object_source)

            if object_type == CodeType.__name__:
                return CodeType(*[cls.get_object(object_source[prop]) for prop in CODE_PROPERTIES])

            if object_type == CellType.__name__:
                return CellType(cls.get_object(object_source))

            if object_type == staticmethod.__name__:
                return staticmethod(cls.get_object(object_source))

            if object_type == classmethod.__name__:
                return classmethod(cls.get_object(object_source))

            if object_type == FunctionType.__name__:
                return cls.__create_function(object_source)

            if object_type == type.__name__:
                return cls.__create_class(object_source)

            else:
                return cls.__create_object(object_source)

    @classmethod
    def __create_function(cls, object_source):
        code = cls.get_object(object_source[CODE])
        global_vars = cls.get_object(object_source[GLOBALS])
        name = cls.get_object(object_source[NAME])
        defaults = cls.get_object(object_source[DEFAULTS])
        closure = cls.get_object(object_source[CLOSURE])

        for key in global_vars:
            if key in code.co_name and key in globals():
                global_vars[key] = globals()[key]

        func = FunctionType(code, global_vars, name, defaults, closure)

        if func.__name__ in global_vars:
            func.__globals__.update({func.__name__: func})

        return func

    @classmethod
    def __create_class(cls, object_source):
        name = cls.get_object(object_source[NAME])
        bases = cls.get_object(object_source[BASES])
        dictionary = object_source[DICT]
        dictionary = {cls.get_object(key): cls.get_object(
            value) for key, value in dictionary.items()}

        cl = type(name, bases, dictionary)

        for attr in cl.__dict__.values():
            if inspect.isroutine(attr):
                if type(attr) in (staticmethod, classmethod):
                    fglobs = attr.__func__.__globals__
                else:
                    fglobs = attr.__globals__

                for gv in fglobs.keys():
                    if gv == cl.__name__:
                        fglobs[gv] = cl

        return cl

    @classmethod
    def __create_object(cls, object_source):
        cl = cls.get_object(object_source[CLASS])
        dictionary = object_source[DICT]
        dictionary = {cls.get_object(key): cls.get_object(
            value) for key, value in dictionary.items()}

        obj = object.__new__(cl)
        obj.__dict__ = dictionary

        return obj

    @classmethod
    def __get_code_dict(cls, obj):
        result = {}

        for key, value in inspect.getmembers(obj):
            if key in CODE_PROPERTIES:
                result[key] = cls.get_dict(value)

        return result

    @classmethod
    def __get_object_dict(cls, obj):
        result = {}

        # Class
        result[CLASS] = cls.get_dict(obj.__class__)

        # Dict
        result[DICT] = cls.__get_obj_dict(obj)

        return result

    @classmethod
    def __get_routine_dict(cls, obj, is_inner_func):
        result = {}

        # Code
        result[CODE] = cls.get_dict(obj.__code__)

        # Global vars
        global_vars = cls.__get_global_vars(obj, is_inner_func)
        result[GLOBALS] = cls.get_dict(global_vars)

        # Name
        result[NAME] = cls.get_dict(obj.__name__)

        # Defaults
        result[DEFAULTS] = cls.get_dict(obj.__defaults__)

        # Closure
        result[CLOSURE] = cls.get_dict(obj.__closure__)

        return result

    # @classmethod
    # def __get_global_vars(cls, func, is_inner_func):
    #     name = func.__name__
    #     global_vars = {}
    #
    #     for var_name in func.__code__.co_names:
    #         if var_name in func.__globals__:
    #             # Module
    #             if type(func.__globals__[var_name]) is ModuleType:
    #                 global_vars[var_name] = func.__globals__[var_name]
    #
    #             # Class
    #             elif inspect.isclass(func.__globals__[var_name]):
    #                 c = func.__globals__[var_name]
    #
    #                 if is_inner_func and name in c.__dict__ and func == c.__dict__[name].__func__:
    #                     global_vars[var_name] = c.__name__
    #                 else:
    #                     global_vars[var_name] = c
    #
    #             elif var_name == func.__code__.co_name:
    #                 global_vars[var_name] = func.__name__
    #
    #             else:
    #                 global_vars[var_name] = func.__globals__[var_name]
    #
    #     return global_vars

    @classmethod
    def __get_global_vars(cls, func, is_inner_func):
        name = func.__name__
        gvars = {}

        for gvar_name in func.__code__.co_names:
            # Separating the variables that the function needs
            if gvar_name in func.__globals__:

                # Module
                if type(func.__globals__[gvar_name]) is ModuleType:
                    gvars[gvar_name] = func.__globals__[gvar_name]

                # Class
                elif inspect.isclass(func.__globals__[gvar_name]):
                    # To prevent recursion, the class in which this method is declared is replaced with the
                    # name of the class. In the future, this name will be replaced by the class type
                    c = func.__globals__[gvar_name]
                    if is_inner_func and name in c.__dict__ and func == c.__dict__[name].__func__:  # !!!!
                        gvars[gvar_name] = c.__name__
                    else:
                        gvars[gvar_name] = c

                # Recursion protection
                elif gvar_name == func.__code__.co_name:
                    gvars[gvar_name] = func.__name__

                else:
                    gvars[gvar_name] = func.__globals__[gvar_name]

        return gvars

    @classmethod
    def __get_obj_dict(cls, obj):
        dictionary = {key: value for key, value in obj.__dict__.items()}
        dictionary2 = {}

        for key, value in dictionary.items():
            if type(value) not in UNIQUE_TYPES:
                if inspect.isroutine(value):
                    dictionary2[cls.get_dict(key)] = cls.get_dict(value, is_inner_func=True)
                else:
                    dictionary2[cls.get_dict(key)] = cls.get_dict(value)

        return dictionary2

    @classmethod
    def __get_class_dict(cls, obj):
        result = {}

        # Name
        result[NAME] = cls.get_dict(obj.__name__)

        # Bases
        result[BASES] = cls.get_dict(tuple(base for base in obj.__bases__ if base != object))

        # Dict
        result[DICT] = cls.__get_obj_dict(obj)

        return result



from .utility_functions import ObjectConverter
from types import NoneType, ModuleType, CodeType, FunctionType, CellType
import re
import regex
from .constants import NULL_LITERAL, INT_PATTERN, FLOAT_PATTERN, BOOL_PATTERN,\
    TRUE_LITERAL, STRING_PATTERN, NULL_PATTERN, VALUE_PATTERN


class MyJson:
    def dumps(self, obj) -> str:
        obj = ObjectConverter.get_dict(obj)

        return self._dumps(obj)

    def dump(self, obj, source_file):
        source_file.write(self.dumps(obj))

    def load(self, source_file):
        return self.loads(source_file.read())

    def _dumps(self, obj) -> str:
        if type(obj) in (int, float):
            return str(obj)

        if type(obj) is bool:
            return str(obj).lower()

        if type(obj) is str:
            return '"' + self.__mask_quotes(obj) + '"'

        if type(obj) is NoneType:
            return NULL_LITERAL

        if type(obj) is list:
            return '[' + ", ".join([self._dumps(item) for item in obj]) + ']'

        if type(obj) is dict:
            return '{' + ", ".join([f"{self._dumps(item[0])}: "
                                    f"{self._dumps(item[1])}" for item in obj.items()]) + '}'
        else:
            raise ValueError

    def loads(self, string: str):
        obj = self._loads(string)
        return ObjectConverter.get_object(obj)

    def _loads(self, string: str):
        string = string.strip()

        match = re.fullmatch(INT_PATTERN, string)
        if match:
            return int(match.group(0))

        match = re.fullmatch(FLOAT_PATTERN, string)
        if match:
            return float(match.group(0))

        match = re.fullmatch(BOOL_PATTERN, string)
        if match:
            return match.group(0) == TRUE_LITERAL

        match = re.fullmatch(STRING_PATTERN, string)
        if match:
            ans = match.group(0)
            ans = self.__unmask_quotes(ans)
            return ans[1:-1]

        match = re.fullmatch(NULL_PATTERN, string)
        if match:
            return None

        # List
        if string[0] == '[' and string[-1] == ']':
            string = string[1:-1]
            matches = regex.findall(VALUE_PATTERN, string)
            return [self._loads(match[0]) for match in matches]

        # Dict
        if string[0] == '{' and string[-1] == '}':
            string = string[1:-1]
            matches = regex.findall(VALUE_PATTERN, string)

            return {self._loads(matches[i][0]):
                    self._loads(matches[i + 1][0]) for i in range(0, len(matches), 2)}

        else:
            raise ValueError

    @staticmethod
    def __mask_quotes(string: str) -> str:
        return string.replace('\\', "\\\\").replace('"', r"\"").replace("'", r"\'")

    @staticmethod
    def __unmask_quotes(string: str) -> str:
        return string.replace('\\\\', "\\").replace(r"\"", '"').replace(r"\'", "'")



import regex
from .utility_functions import ObjectConverter
from types import NoneType
from .constants import FIRST_XML_ELEMENT_PATTERN, XML_ELEMENT_PATTERN, KEY_GROUP_NAME, VALUE_GROUP_NAME, XML_SCHEME_SOURCE


class MyXml:
    def dumps(self, obj) -> str:
        obj = ObjectConverter.get_dict(obj)
        return self._dumps(obj, is_first=True)

    def dump(self, obj, source_file):
        source_file.write(self.dumps(obj))

    def load(self, source_file):
        return self.loads(source_file.read())

    def _dumps(self, obj, is_first=False) -> str:
        if type(obj) in (int, float, bool, NoneType):
            return self.__create_xml_element(type(obj).__name__, str(obj), is_first)

        if type(obj) is str:
            data = self.__mask_symbols(obj)
            return self.__create_xml_element(str.__name__, data, is_first)

        if type(obj) is list:
            data = ''.join([self._dumps(o) for o in obj])
            return self.__create_xml_element(list.__name__, data, is_first)

        if type(obj) is dict:
            data = ''.join(
                [f"{self._dumps(item[0])}{self._dumps(item[1])}" for item in obj.items()])
            return self.__create_xml_element(dict.__name__, data, is_first)

        else:
            raise ValueError

    def loads(self, string: str):
        obj = self._loads(string, is_first=True)
        return ObjectConverter.get_object(obj)

    def _loads(self, string: str, is_first=False):
        string = string.strip()
        xml_element_pattern = FIRST_XML_ELEMENT_PATTERN if is_first else XML_ELEMENT_PATTERN

        match = regex.fullmatch(xml_element_pattern, string)

        if not match:
            raise ValueError

        key = match.group(KEY_GROUP_NAME)
        value = match.group(VALUE_GROUP_NAME)

        if key == int.__name__:
            return int(value)

        if key == float.__name__:
            return float(value)

        if key == bool.__name__:
            return value == str(True)

        if key == str.__name__:
            return self.__unmask_symbols(value)

        if key == NoneType.__name__:
            return None

        if key == list.__name__:
            matches = regex.findall(XML_ELEMENT_PATTERN, value)
            return [self._loads(match[0]) for match in matches]

        if key == dict.__name__:
            matches = regex.findall(XML_ELEMENT_PATTERN, value)
            return {self._loads(matches[i][0]):
                    self._loads(matches[i + 1][0]) for i in range(0, len(matches), 2)}
        else:
            raise ValueError

    def __create_xml_element(self, name: str, data: str, is_first=False):
        if is_first:
            return f"<{name} {XML_SCHEME_SOURCE}>{data}</{name}>"
        else:
            return f"<{name}>{data}</{name}>"

    @staticmethod
    def __mask_symbols(string: str) -> str:
        return string.replace('&', "&amp;").replace('<', "&lt;").replace('>', "&gt;"). \
            replace('"', "&quot;").replace("'", "&apos;")

    @staticmethod
    def __unmask_symbols(string: str) -> str:
        return string.replace("&amp;", '&').replace("&lt;", '<').replace("&gt;", '>'). \
            replace("&quot;", '"').replace("&apos;", "'")



import types
from enum import Enum

# KEY WORDS
TYPE = "type"
SOURCE = "source"

CODE = "__code__"
GLOBALS = types.FunctionType.__globals__.__name__
NAME = "__name__"
DEFAULTS = "__defaults__"
CLOSURE = types.FunctionType.__closure__.__name__

BASES = "__bases__"
DICT = "__dict__"

CLASS = "__class__"
OBJECT = "object"


CODE_PROPERTIES = [prop.__name__ for prop in [
    types.CodeType.co_argcount,
    types.CodeType.co_posonlyargcount,
    types.CodeType.co_kwonlyargcount,
    types.CodeType.co_nlocals,
    types.CodeType.co_stacksize,
    types.CodeType.co_flags,
    types.CodeType.co_code,
    types.CodeType.co_consts,
    types.CodeType.co_names,
    types.CodeType.co_varnames,
    types.CodeType.co_filename,
    types.CodeType.co_name,
    types.CodeType.co_firstlineno,
    types.CodeType.co_lnotab,
    types.CodeType.co_freevars,
    types.CodeType.co_cellvars]
]


UNIQUE_TYPES = [
    types.MappingProxyType,
    types.WrapperDescriptorType,
    types.MemberDescriptorType,
    types.GetSetDescriptorType,
    types.BuiltinFunctionType
]

# JSON

INF_LITERAL = str(1E1000)
NAN_LITERAL = str(1E1000 / 1E1000)

TRUE_LITERAL = "true"
FALSE_LITERAL = "false"

NULL_LITERAL = "null"

INT_PATTERN = fr"[+-]?\d+"
FLOAT_PATTERN = fr"(?:[+-]?\d+(?:\.\d+)?(?:e[+-]?\d+)?|[+-]?{INF_LITERAL}\b|{NAN_LITERAL}\b)"
BOOL_PATTERN = fr"({TRUE_LITERAL}|{FALSE_LITERAL})\b"
STRING_PATTERN = fr"\"(?:(?:\\\")|[^\"])*\""
NULL_PATTERN = fr"\b{NULL_LITERAL}\b"

ELEMENTARY_TYPES_PATTERN = fr"{FLOAT_PATTERN}|{INT_PATTERN}|{BOOL_PATTERN}|{STRING_PATTERN}|{NULL_PATTERN}"

# This regex use recursive statements to be able to capture nested lists and objects.
ARRAY_PATTERN = r"\[(?R)?(?:,(?R))*\]"
OBJECT_PATTERN = r"\{(?:(?R):(?R))?(?:,(?R):(?R))*\}"

VALUE_PATTERN = fr"\s*({ELEMENTARY_TYPES_PATTERN}|" + \
                fr"{ARRAY_PATTERN}|{OBJECT_PATTERN})\s*"


# XML 

KEY_GROUP_NAME = "key"
VALUE_GROUP_NAME = "value"

XML_SCHEME_SOURCE = "xmlns:xsi=\"http://www.w3.org/2001/XMLSchema-instance\" " + \
                    "xmlns:xsd=\"http://www.w3.org/2001/XMLSchema\""

XML_SCHEME_PATTERN = "xmlns:xsi=\"http://www\.w3\.org/2001/XMLSchema-instance\" " + \
    "xmlns:xsd=\"http://www\.w3\.org/2001/XMLSchema\""

ELEMENTARY_NAMES_PATTERN = "int|float|bool|str|NoneType|list|dict"

XML_ELEMENT_PATTERN = fr"(\<(?P<{KEY_GROUP_NAME}>{ELEMENTARY_NAMES_PATTERN})\>" + \
    fr"(?P<{VALUE_GROUP_NAME}>([^<>]*)|(?R)+)\</(?:{ELEMENTARY_NAMES_PATTERN})\>)"

FIRST_XML_ELEMENT_PATTERN = fr"(\<(?P<{KEY_GROUP_NAME}>{ELEMENTARY_NAMES_PATTERN})\s*({XML_SCHEME_PATTERN})?\>" + \
                            fr"(?P<{VALUE_GROUP_NAME}>([^<>]*)|(?R)+)\</(?:{ELEMENTARY_NAMES_PATTERN})\>)"


# ENUM FOR 
class Serializer(Enum):
    Json = 1,
    Xml = 2



from setuptools import setup

setup(
    name="My_own_serializer",
    version="0.0",
    description="Library for converting objects to xml and json format",
    url="https://github.com/andrw04/Python-labs/tree/lab3",
    author="Andrei Sivy",
    author_email="sivyandrey2021@gmail.com",
    license="MIT",
    classifiers=[
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.10",
        "Operating System :: OS Independent"
    ],
    packages=["serializer"],
    include_package_data=True,
    install_requires=["regex"]
)